unit Unit1;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, MemoEx, StdCtrls, ExtCtrls, CheckLst, ComCtrls, Menus;

type
  TForm1 = class(TForm)
    Panel1: TPanel;
    Splitter1: TSplitter;
    Label1: TLabel;
    MemoEx1: TMemoEx;
    cblSettings: TCheckListBox;
    Label2: TLabel;
    edRightMargin: TEdit;
    Label3: TLabel;
    edAutoIndentSize: TEdit;
    Label4: TLabel;
    edIndentSize: TEdit;
    Label5: TLabel;
    edTabSize: TEdit;
    btnApplySettings: TButton;
    Label6: TLabel;
    edGutterWidth: TEdit;
    MemoExStatus: TStatusBar;
    MemoPopup: TPopupMenu;
    Cut1: TMenuItem;
    Copy1: TMenuItem;
    Paste1: TMenuItem;
    Delete1: TMenuItem;
    SelectAll1: TMenuItem;
    N1: TMenuItem;
    N2: TMenuItem;
    Undo1: TMenuItem;
    SaveDialog: TSaveDialog;
    Image1: TImage;
    procedure FormShow(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure btnApplySettingsClick(Sender: TObject);
    procedure MemoEx1SelectionChange(Sender: TObject);
    procedure MemoEx1ChangeClipboardState(Sender: TObject;
      const CanPaste: Boolean);
    procedure MemoEx1ChangeStatus(Sender: TObject);
    procedure cblSettingsEnter(Sender: TObject);
    procedure cblSettingsExit(Sender: TObject);
    procedure MemoEx1GetLineAttr(Sender: TObject; const Line: String;
      Index: Integer; const SelAttrs: TSelAttrs; var Attrs: TLineAttrs);
    procedure MemoEx1MouseOver(Sender: TObject; WordStyle: Word;
      var _Cursor: TCursor);
    procedure MemoEx1WordClick(Sender: TObject; const WordText: String;
      WordStyle: Word);
    procedure Cut1Click(Sender: TObject);
    procedure Copy1Click(Sender: TObject);
    procedure Paste1Click(Sender: TObject);
    procedure Delete1Click(Sender: TObject);
    procedure Undo1Click(Sender: TObject);
    procedure SelectAll1Click(Sender: TObject);
    function MemoEx1BlockOperation(Sender: TObject; MacroID: Integer;
      const Text: String): String;
    procedure MemoEx1BreakLine(Sender: TObject; const Original: String;
      var _New: String);
    procedure MemoEx1ConcatLine(Sender: TObject; const Original: String;
      var _New: String);
    function MemoEx1InsertBlock(Sender: TObject;
      var Text: String): Boolean;
    function MemoEx1InsertMacro(Sender: TObject; MacroID: Integer): String;
    procedure MemoEx1SaveBlock(Sender: TObject; const Text: String);
    procedure MemoEx1TextInsert(Sender: TObject; var Text: String);
    procedure MemoEx1PaintGutter(Sender: TObject; Canvas: TCanvas);
  private
    { Private declarations }
    SyntaxHighlight: boolean;
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

uses
  Math;

{$R *.dfm}

const
  _STYLE_NORMAL         = 0;                //  normal text
  _STYLE_INVISIBLE      = 1;                //  invisible text
  _STYLE_BOLD           = 2;                //  bold text
  _STYLE_ITALIC         = 3;                //  italic
  _STYLE_UNDERLINE      = 4;                //  underlined text
  _STYLE_SPECIAL        = 5;                //  special text
  _STYLE_HOTSPOT        = 6;                //  hotspot

  _STYLE_FIRST          = _STYLE_NORMAL;
  _STYLE_LAST           = _STYLE_HOTSPOT;

var
  Styles: array [_STYLE_FIRST.._STYLE_LAST] of TLineAttr;

{
  Fill styles with default attributes.
}
procedure SetDefaultTextStyles;

  procedure SetDefaultTextStyle(ATextStyle: integer; FC, BC: TColor; Style: TFontStyles; ExStyle: byte);
  begin
    Styles[ATextStyle].FC := FC;
    Styles[ATextStyle].BC := BC;
    Styles[ATextStyle].Style := Style;
    Styles[ATextStyle].ex_style := ExStyle;
  end;

begin
  SetDefaultTextStyle(_STYLE_NORMAL, clWindowText, clWindow, [], 0);
  SetDefaultTextStyle(_STYLE_INVISIBLE, clWindow, clWindow, [], 0);
  SetDefaultTextStyle(_STYLE_BOLD, clWindowText, clWindow, [fsBold], 0);
  SetDefaultTextStyle(_STYLE_ITALIC, clWindowText, clWindow, [fsItalic], 0);
  SetDefaultTextStyle(_STYLE_UNDERLINE, clWindowText, clWindow, [fsUnderline], 0);
  SetDefaultTextStyle(_STYLE_SPECIAL, clMaroon, clWindow, [fsBold], 0);
  SetDefaultTextStyle(_STYLE_HOTSPOT, clBlue, clWindow, [fsUnderline], _STYLE_HOTSPOT);
end;

procedure TForm1.FormShow(Sender: TObject);
begin
  with cblSettings do
  begin
    Checked[0] := true;       //  word wrap
    Checked[1] := true;       //  read-only
    Checked[2] := true;       //  highlight
    Checked[3] := true;       //  strip invisible
    Checked[4] := true;       //  cursor beyond end of line
    Checked[5] := false;      //  cursor beyond end of file
    Checked[6] := true;       //  auto indent
    Checked[7] := true;       //  smart auto indent
    Checked[8] := false;      //  smart tab
    Checked[9] := true;       //  simple begin line
    Checked[10] := true;      //  right margin visible
    Checked[11] := true;      //  enabled
    Checked[12] := true;      //  completion enabled
    Checked[13] := false;     //  unicode output
    Checked[14] := false;     //  auto word wrap
  end;
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  //  enable syntax highlighting by default
  SyntaxHighlight := true;
  SetDefaultTextStyles;
end;

procedure TForm1.btnApplySettingsClick(Sender: TObject);
var
  NeedRedraw: boolean;
begin
  //  apply all selected settings to control
  with cblSettings do
  begin
    NeedRedraw := (MemoEx1.ReadOnly <> Checked[1]) or (SyntaxHighlight <> Checked[2])
                    or (MemoEx1.StripInvisible <> Checked[3]);
    MemoEx1.WordWrap := Checked[0];
    MemoEx1.ReadOnly := Checked[1];
    SyntaxHighlight := Checked[2];
    MemoEx1.StripInvisible := Checked[3];
    MemoEx1.CursorBeyondEOL := Checked[4];
    MemoEx1.CursorBeyondEOF := Checked[5];
    MemoEx1.AutoIndent := Checked[6];
    MemoEx1.SmartAutoIndent := Checked[7];
    MemoEx1.SmartTab := Checked[8];
    MemoEx1.SimpleBeginLine := Checked[9];
    MemoEx1.RightMarginVisible := Checked[10];
    MemoEx1.Enabled := Checked[11];
    MemoEx1.Completion.Enabled := Checked[12];
    MemoEx1.RightMargin := StrToIntDef(edRightMargin.Text, 80);
    MemoEx1.AutoIndentSize := StrToIntDef(edAutoIndentSize.Text, 0);
    MemoEx1.IndentSize := StrToIntDef(edIndentSize.Text, 4);
    MemoEx1.TabSize := StrToIntDef(edTabSize.Text, 4);
    MemoEx1.GutterWidth := StrToIntDef(edGutterWidth.Text, 30);
  end;
  if NeedRedraw then MemoEx1.Invalidate2;
  MemoEx1.SetFocus;
end;

procedure TForm1.MemoEx1SelectionChange(Sender: TObject);
var
  s: string;
begin
  //  perform interface updates depending on text selection changes
  if MemoEx1.SelectedText then s := 'selection present'
  else s := 'selection is not present';
  MemoExStatus.Panels[3].Text := s;
  Copy1.Enabled := MemoEx1.SelectedText;
  Cut1.Enabled := MemoEx1.SelectedText;
  Delete1.Enabled := MemoEx1.SelectedText;
end;

procedure TForm1.MemoEx1ChangeClipboardState(Sender: TObject;
  const CanPaste: Boolean);
var
  s: string;
begin
  //  perform interface updates depending on clipboard contents
  if CanPaste then s := 'can paste text'
  else s := 'no text to paste';
  MemoExStatus.Panels[4].Text := s;
  Paste1.Enabled := CanPaste;
end;

procedure TForm1.MemoEx1ChangeStatus(Sender: TObject);
var
  s: string;
begin
  //  perform interface updates depending on status changes
  MemoExStatus.Panels[0].Text := Format('%d: %d', [MemoEx1.CaretX + 1, MemoEx1.CaretY + 1]);
  if MemoEx1.Recording then s := 'macro recording'
  else
    if MemoEx1.InsertMode then s := 'insert'
    else s := 'overwrite';
  MemoExStatus.Panels[1].Text := s;
  if MemoEx1.CanUndo then s := 'can undo'
  else s := 'no operations to undo';
  MemoExStatus.Panels[2].Text := s;
  Undo1.Enabled := MemoEx1.SelectedText;
end;

procedure TForm1.cblSettingsEnter(Sender: TObject);
begin
  btnApplySettings.Default := true;
end;

procedure TForm1.cblSettingsExit(Sender: TObject);
begin
  btnApplySettings.Default := false;
end;

procedure TForm1.MemoEx1GetLineAttr(Sender: TObject; const Line: String;
  Index: Integer; const SelAttrs: TSelAttrs; var Attrs: TLineAttrs);

  procedure SetTextStyle(const AIndex: integer; const AStyle: byte);
  begin
    //  Not changed Attr[AIndex].BC always defaults to MemoEx.Color property.
    Attrs[AIndex] := Styles[AStyle];
  end;

  {
    Search for lines which starts with '>' -- special lines.
  }
  function SpecialsFound: boolean;
  var
    i: integer;
  begin
    Result := false;
    for i := 1 to length(Line) do
      if i < 10 then
        if ord(Line[i]) = 62 then
        begin
          //  The line starts with '>' is what we need.
          Result := true;
          break;
        end
          else
      else break;
    if Result then
      for i := 1 to length(Line) do SetTextStyle(i - 1, _STYLE_SPECIAL);
  end;

var
  i, j, k, x1, x2: integer;
begin
  //  If we do not need to highlight the text, we leave.
  if not SyntaxHighlight then exit;
  //  Let's find special lines.
  if SpecialsFound then exit;
  j := _STYLE_NORMAL;                     //  normal style
  x1 := 0;
  x2 := 0;
  for i := 1 to length(Line) do
  begin
    case ord(Line[i]) of
      32:                                 //  space resets any style to normal
        begin
          k := _STYLE_NORMAL;
          j := _STYLE_NORMAL;
          x1 := 0;
          x2 := 0;
        end;
      42:                                 //  bold (*text*)
        k := _STYLE_BOLD;
      47:                                 //  italic (/text/)
        k := _STYLE_ITALIC;
      95:                                 //  underline (_text_)
        k := _STYLE_UNDERLINE;
      126:                                //  hotspot (~text~)
        k := _STYLE_HOTSPOT;
      else                                //  normal
        k := _STYLE_NORMAL;
    end;
    if k > _STYLE_NORMAL then
    begin
      if (j = _STYLE_NORMAL) or (j <> k) then
      begin
        x1 := i;
        j := k;
        x2 := 0;
      end
      else
        if j = k then x2 := i;
      if (x1 > 0) and (x2 - 1 >= x1 + 1) then
      begin
        //  Highlight the text.
        for k := x1 + 1 to x2 - 1 do SetTextStyle(k - 1, j);
        if MemoEx1.ReadOnly then
        begin
          SetTextStyle(x1 - 1, _STYLE_INVISIBLE);
          SetTextStyle(x2 - 1, _STYLE_INVISIBLE);
        end;
        j := _STYLE_NORMAL;
        x1 := 0;
        x2 := 0;
      end
      else
        if (x1 > 0) and (x2 > 0) then
        begin
          //  Oops. We shouldn't reach this place.
          x1 := 0;
          x2 := 0;
        end;
    end;
  end;
end;

procedure TForm1.MemoEx1MouseOver(Sender: TObject; WordStyle: Word;
  var _Cursor: TCursor);
begin
  //  mouse was moved over word, let's check the word style and update a cursor
  if WordStyle = _STYLE_HOTSPOT then _Cursor := crHandPoint;
end;

procedure TForm1.MemoEx1WordClick(Sender: TObject; const WordText: String;
  WordStyle: Word);
begin
  //  user has clicked a left mouse button, let's check the word style
  if WordStyle = _STYLE_HOTSPOT then
    MessageDlg(Format('"%s" clicked!', [WordText]), mtInformation, [mbOk], 0);
end;

procedure TForm1.Cut1Click(Sender: TObject);
begin
  //  cut to clipboard, from pop-up menu
  MemoEx1.CutToClipboard;
end;

procedure TForm1.Copy1Click(Sender: TObject);
begin
  //  copy to clipboard, from pop-up menu
  MemoEx1.CopyToClipboard;
end;

procedure TForm1.Paste1Click(Sender: TObject);
begin
  //  paste from clipboard, from pop-up menu
  MemoEx1.PasteFromClipboard;
end;

procedure TForm1.Delete1Click(Sender: TObject);
begin
  //  delete a selected text, from pop-up menu
  MemoEx1.ClearSelection;
end;

procedure TForm1.Undo1Click(Sender: TObject);
begin
  //  undo, from pop-up menu
  MemoEx1.Undo;
end;

procedure TForm1.SelectAll1Click(Sender: TObject);
begin
  //  select all text from pop-up menu
  MemoEx1.SelectAll;
end;

function TForm1.MemoEx1BlockOperation(Sender: TObject; MacroID: Integer;
  const Text: String): String;
begin
  //  user-defined block operation (Ctrl-B,{A-Z})
  MessageDlg(Format('Process with: "%s"', [Text]), mtInformation, [mbOk], 0);
  Result := Text;
end;

procedure TForm1.MemoEx1BreakLine(Sender: TObject; const Original: String;
  var _New: String);
begin
  //  user-processed line break
  if pos('>', TrimLeft(Original)) = 1 then _New := '> ' + _New;
end;

procedure TForm1.MemoEx1ConcatLine(Sender: TObject; const Original: String;
  var _New: String);
var
  //  user-processed line concatenation
  i: integer;
begin
  i := pos('>', TrimLeft(_New));
  if pos('>', TrimLeft(Original)) = i then
  begin
    _New := TrimLeft(_New);
    System.Delete(_New, 1, i + 1);
  end;
end;

function TForm1.MemoEx1InsertBlock(Sender: TObject;
  var Text: String): Boolean;
begin
  //  insert block from the external source (Ctrl-K,R)
  Text := InputBox('Insert block', 'Text to insert from "external resource"', '');
  Result := Text <> '';
end;

function TForm1.MemoEx1InsertMacro(Sender: TObject;
  MacroID: Integer): String;
begin
  //  insert user-defined macro (Ctrl-S,{A-Z})
  Result := Format('Macro%d', [MacroID]);
end;

procedure TForm1.MemoEx1SaveBlock(Sender: TObject; const Text: String);
var
  F: TextFile;
begin
  //  save block to file (Ctrl-K,W)
  if SaveDialog.Execute then
  begin
    AssignFile(F, SaveDialog.FileName);
    Rewrite(F);
    Write(F, Text);
    CloseFile(F);
  end;
end;

procedure TForm1.MemoEx1TextInsert(Sender: TObject; var Text: String);
begin
  //  user-defined handler for any text insertions
  MessageDlg(Format('Text "%s" will be inserted', [Text]), mtInformation, [mbOk], 0);
end;

procedure TForm1.MemoEx1PaintGutter(Sender: TObject; Canvas: TCanvas);
var
  i, j: integer;
  rgn: HRGN;
begin
  //  gutter painter
  j := Max(1, MemoEx1.Height div 100);
  rgn := CreateRectRgn(0, 0, MemoEx1.GutterWidth - 2, MemoEx1.Height);
  SelectClipRgn(Canvas.Handle, rgn);
  for i := 1 to j + 1 do
    Canvas.Draw(-30, (i - 1) * 100, Image1.Picture.Graphic);
  SelectClipRgn(Canvas.Handle, 0);
  DeleteObject(rgn);
end;

end.
